(** Copyright (c) 2018 by Swirlds, Inc.
   Implemented by Karl Crary.
*)

Require Import Nat.
Require Import Bool.
Require Import List.
Require Import Omega.

Require Import Tact.
Require Import Relation.
Require Import Decide.
Require Import Weight.
Require Import Calculate.
Require Import Majority.
Require Import HashgraphFacts.
Require Import Sees.
Require Import Round.
Require Import Progress.
Require Import Vote.
Require Import Decision.
Require Import Consensus.


Definition famous (s : sample) (W : world) (x : event) : Prop :=
  exists y, member W y /\ decision s x y true.


Definition fwitness s W i x :=
  rwitness i x
  /\ famous s W x.


Lemma extends_famous :
   forall s W W' x,
     extends W W'
     -> famous s W x
     -> famous s W' x.
Proof.
intros s W W' x Hextends Hfamous.
destruct Hfamous as (y & Wy & Hdecision).
exists y.
split; auto.
Qed.


Lemma fame_reaches_every_world :
  forall s W W' x,
    extends W W'
    -> nonterminating W
    -> famous s W' x
    -> famous s W x.
Proof.
intros s W W' x Hw Hnonterm Hfamous.
destruct Hfamous as (y & Hsy & Hdecision).
so Hdecision as (i & j & _ & _ & Hwitx & Hwity & Hij & _).
assert (exists k, j < k /\ (k - i) mod coin_freq <> 0) as (k & Hjk & Hnocoin).
  {
  so (eq_nat_dec ((S j - i) mod coin_freq) 0) as [Heq | Hneq].
    {
    exists (2 + j).
    split; try omega.
    replace (2 + j - i) with (S (S j - i)) by omega.
    apply succ_no_coin; auto.
    }

    {
    exists (S j).
    auto.
    }
  }
so (all_rounds_inhabited _ k Hnonterm) as (z & Hwitz & Wz).
exploit (decisions_propagate _#7 Hsy (Hwitx andel) (Hwity andel) Hdecision k z) as Hdecision'; auto.
exists z.
auto.
Qed.


Lemma famous_decide_finite :
  forall s W x,
    finite (member W)
    -> decidable (famous s W x).
Proof.
intros s W x Hfin.
apply (dec_exists_finite _ _ (member W)); auto.
  {
  intros y Hy.
  destruct Hy; auto.
  }
intros y.
apply dec_and.
  {
  apply finite_decide; auto.
  exact eq_event_decide.
  }

  {
  apply decision_decide.
  }
Qed.


Lemma famous_decide_nonterm :
  forall s W x,
    extends W (global s)
    -> nonterminating W
    -> member W x
    -> witness x
    -> decidable (famous s W x).
Proof.
intros s W x HsW Hnonterm Wx Hwitx.
so (fame_consensus _ _ (HsW _ Wx) Hwitx) as (y & b & Hsy & Hdecision).
destruct b.
  {
  left.
  apply (fame_reaches_every_world _#4 HsW Hnonterm).
  exists y.
  auto.
  }

  {
  right.
  intros (z & Wz & Hdecision').
  so (decision_consistent _#7 Hsy (HsW _ Wz) Hdecision Hdecision') as H.
  discriminate H.
  }
Qed.


Lemma elector_cardinality :
  forall i y,
    exists n, weight creatorstake (fun x => elector i x y) n. 
Proof.
intros i y.
apply finite_weight.
  { exact eq_event_decide. }

  {
  apply (finite_subset _ _ (fun w => w @ y)).
    {
    intros x (_ & Hx).
    apply stsees_impl_strict_ancestor; auto.
    }
    
    {
    intros x.
    apply dec_and.
      { apply rwitness_decide. }

      { apply stsees_decide. }
    }

    {
    apply strict_ancestor_finite.
    }
  }
Qed.


Definition backward (i : nat) (P : event -> Prop) (x : event) : Prop :=
  rwitness i x
  /\ 
  exists y, P y /\ stsees x y.


Lemma backward_weight :
  forall W i j (P : event -> Prop),
    i < j
    -> (forall x, P x -> member W x /\ rwitness j x)
    -> (exists x, P x)
    -> finite P
    -> exists n,
         weight creatorstake (backward i P) n
         /\ two_thirds total_stake < n
         /\ n <= total_stake
         /\ (forall x, backward i P x -> member W x).
Proof.
intros W i j P Hij Hallwit (w & Hw) Hfinite.
exploit (finite_weight _ creatorstake (backward i P)) as H.
  { exact eq_event_decide. }

  {
  apply (finite_equiv _ (fun x => exists y, P y /\ rwitness i x /\ stsees x y)).
    {
    intros x.
    split.
      {
      intros (y & Hy & Hwit & Hxy).
      split; auto.
      exists y.
      auto.
      }

      {
      intros (Hwit & y & Hy & Hxy).
      exists y.
      auto.
      }
    }

    {
    apply finite_exists; auto.
    intros x Hx.
    apply (finite_subset _ _ (fun y => y @ x)).
      {
      intros y (_ & Hy).
      apply stsees_impl_strict_ancestor; auto.
      }
      
      {
      intros y.
      apply dec_and.
        { apply rwitness_decide. }

        { apply stsees_decide. }
      }

      {
      apply strict_ancestor_finite.
      }
    }
  }
destruct H as (n & Hcard).
exists n.
do2 3 split; auto.
  {
  so (Hallwit _ Hw ander) as Hwitw.
  assert (S i <= j) as Hij' by omega.
  so (ancestor_witness _#3 Hij' (Hwitw andel)) as (v & Hvw & Hwitv).
  so (elector_cardinality i v) as (m & Hcard').
  refine (lt_le_trans _#3 (elector_minimum _#3 (Hwitv andel) Hcard') _).
  refine (weight_incl _#5 _ _ Hcard' Hcard).
  intros x Hx.
  destruct Hx as (Hwitx & Hxv).
  split; auto.
  exists w.
  split; auto.
  eapply stsees_ancestor_trans; eauto.
  }

  {
  apply (weight_map_inj _ _ creatorstake stake (backward i P) every creator); auto.
    { unfold every; auto. }

    {
    intros x y Hx Hy Heqcr.
    destruct Hx as (Hwitx & x' & Px' & Hxx).
    destruct Hy as (Hwity & y' & Py' & Hyy).
    eapply (stseen_witness_unique W); eauto.
      { exact (Hallwit _ Px' andel). }

      { exact (Hallwit _ Py' andel). }
    }

    { exact the_total_stake. }
  }

  {
  intros x Hx.
  destruct Hx as (_ & x' & Px' & Hxx).
  so (Hallwit _ Px' andel).
  eapply world_closed; eauto.
  apply stsees_impl_ancestor; auto.
  }
Qed.


(** Lemma 5.17 *)
Lemma famous_witness_exists :
  forall s i x,
    rwitness (2 + first_regular + i) x
    -> exists y,
         rwitness i y
         /\ decision s y x true.
Proof.
intros s i w Hwitw.
so (event_world w) as (W & Ww).
set (S3 := fun z => z = w).
set (S2 := backward (S first_regular + i) S3).
set (S1 := backward (first_regular + i) S2).
set (S0 := backward i S1).
exploit (backward_weight W (S first_regular + i) (2 + first_regular + i) S3) as H; auto.
  {
  intros z Hz.
  subst S3.
  destruct Hz; auto.
  }

  {
  exists w.
  subst S3.
  cbn.
  reflexivity.
  }

  {
  eapply (weight_finite _ creatorstake).
  refine (weight_iff _#5 _ (weight_singleton _ _ w)).
  intro x.
  unfold S3.
  split; auto.
  }
fold S2 in H.
destruct H as (n2 & Hcard2 & Hlt2 & Hle2 & WS2).
exploit (backward_weight W (first_regular + i) (S first_regular + i) S2) as H; auto.
  {
  intros x Hx.
  split; auto.
  destruct Hx; auto.
  }

  {
  eapply weight_nonempty; eauto.
  omega.
  }

  {
  eapply weight_finite; eauto.
  }
fold S1 in H.
destruct H as (n1 & Hcard1 & Hlt1 & Hle1 & WS1).
exploit (backward_weight W i (first_regular + i) S1) as H; auto.
  { 
  so first_regular_min.
  omega.
  }

  {
  intros x Hx.
  split; auto.
  destruct Hx; auto.
  }

  {
  eapply weight_nonempty; eauto.
  omega.
  }

  {
  eapply weight_finite; eauto.
  }
fold S0 in H.
destruct H as (n0 & Hcard0 & Hlt0 & Hle0 & _).
exploit (weight_tabulate _ _ creatorstake creatorstake S1 (fun y x => S0 x /\ stsees x y) n1 (S (two_thirds total_stake))) as H; auto.
  {
  intros x Hx.
  exploit (backward_weight W i (first_regular + i) (fun y => y = x)) as H; auto.
    { 
    so first_regular_min.
    omega.
    }

    {
    intros ? ->.
    split; auto.
    destruct Hx; auto.
    }
  
    {
    exists x; auto.
    }

    {
    eapply (weight_finite _ creatorstake).
    refine (weight_iff _#5 _ (weight_singleton _ _ x)).
    intro.
    split; auto.
    }
  destruct H as (n' & Hcard & Hlt & _).
  exists n'.
  split; [| omega].
  eapply weight_iff; eauto.
  intro y.
  split.
    {
    intros Hy.
    destruct Hy as (Hwity & ? & -> & Hyx).
    split; auto.
    split; auto.
    exists x.
    auto.
    }

    {
    intros (Hy & Hyx).
    destruct Hy as (Hwity & z & Hz & _).
    split; auto.
    exists x.
    auto.
    }
  }
destruct H as (n & Hcard & Hlen).
match type of Hcard with
| weight _ ?P _ => set (S01 := P)
end.
fold S01 in Hcard.
assert (n0 * two_thirds n1 < n) as Hlen'.
  {
  exploit (two_thirds_shift n0 n1) as Hlt; [omega |].
  refine (lt_le_trans _#3 Hlt _).
  refine (le_trans _#3 _ Hlen).
  apply Nat.mul_le_mono_l.
  apply le_n_S.
  apply two_thirds_mono.
  auto.
  }
exploit (pigeonhole_weight _ _ (fun xy => creatorstake (fst xy)) creatorstake S01 S0 snd n n0 (two_thirds n1)) as H; auto.
  {
  exact eq_event_decide.
  }

  {
  intros xy Hxy.
  destruct Hxy as (? & ? & ?).
  auto.
  }
destruct H as (x & n' & Hx & Hcardx & Hlt).
so Hx as (Hwitx, _).
assert (weight creatorstake (fun y => S1 y /\ stsees x y) n') as Hcardx'.
  {
  refine (weight_iff _#5 _ (weight_image _#5 fst _ _ _ Hcardx)); auto.
    {
    intros y.
    split.
      {
      intros (yx & -> & Hyx & <-).
      destruct yx as (y, x).
      cbn.
      unfold S01 in Hyx; cbn in Hyx.
      destruct_all Hyx; auto.
      }
      
      {
      intros (Hy & Hxy).
      exists (y, x).
      cbn.
      do2 2 split; auto.
      unfold S01; cbn.
      auto.
      }
    }

    {
    intros yx yx' (Hyx, Heqsnd) (Hyx', Heqsnd') Heqfst.
    destruct yx; destruct yx'.
    f_equal; subst; auto.
    }
  }
exploit (weight_subset _ creatorstake (fun y => S1 y /\ ~ stsees x y) S1 n1) as H; auto.
  {
  intros y Hy.
  so (stsees_decide x y) as [Hxy | Hnxy].
    {
    right.
    intro H.
    destruct H; contradiction.
    }

    {
    left.
    auto.
    }
  }

  {
  intros y Hy.
  destruct Hy; auto.
  }
destruct H as (k & Hcardk & _).
assert (k < one_third total_stake) as Hltk.
  {
  so (lt_dec k (one_third total_stake)) as [| Hnlt]; auto.
  exfalso.
  assert (k >= one_third total_stake) as Hgek by omega; clear Hnlt.
  assert (superminority creatorstake (fun y => S1 y /\ ~ stsees x y) S1) as Hmin.
    {
    exists k, n1.
    do2 3 split; auto.
      {
      intros y (H & _); auto.
      }

      {
      so (one_third_mono _ _ Hle1).
      omega.
      }
    }
  assert (supermajority creatorstake (fun y => S1 y /\ stsees x y) S1) as Hmaj.
    {
    exists n', n1.
    do2 3 split; auto.
    intros y (H & _); auto.
    }
  so (superminority_supermajority_intersect _#5 eq_event_decide Hmin Hmaj) as (y & Hy & Hy').
  destruct Hy as (_ & Hy).
  destruct Hy' as (_ & Hy').
  destruct Hy.
  auto.
  }
exists x.
split; auto.
assert (forall y, S2 y -> vote s x y true) as Hunan.
  {
  intros y Hy.
  so Hy as (Hwity & _).
  exploit (election_defined' s x i (first_regular + i) y) as H; auto.
    { 
    so first_regular_min.
    omega.
    }
  destruct H as (t & f & Helection).
  so (destruct_election _#6 Helection) as (Hcardtf & Hcardt & Hcardf).
  assert (f <= total_stake / 3) as Hordf.
    {
    eassert _; [refine (weight_incl _#6 _ Hcardf Hcardk) |].
      {
      intros z ((Hwitz & Hzy) & Hvote).
      split.
        {
        split; auto.
        exists y.
        auto.
        }

        {
        intros Hxz.
        assert (vote s x z true) as Hvote'.
          {
          eapply vote_first; eauto.
            {
            so first_regular_min.
            omega.
            }

            {
            omega.
            }
          cbn.
          split; auto.
          intros _.
          apply stsees_impl_strict_ancestor; auto.
          }
        so (vote_fun _#5 Hvote Hvote') as Heq.
        discriminate Heq.
        }
      }
    so (div3_one_third total_stake).
    omega.
    }
  so (elector_minimum _#3 (Hwity andel) Hcardtf) as Hordtf.
  assert (t >= f) as Hord.
    {
    so (Nat.div_mul_le total_stake 3 2 (Nat.neq_succ_0 _)).
    unfold two_thirds in Hordtf.
    omega.
    }
  eapply vote_regular; eauto.
    {
    omega.
    }
    
    {
    so coin_freq_min.
    rewrite -> Nat.mod_small; omega.
    }
  }
exploit (election_defined' s x i (S first_regular + i) w) as H; auto.
  { omega. }
destruct H as (t & f & Helection).
exists i, (2 + first_regular + i), t, f.
do2 5 split; auto.
  { omega. }

  { 
  so coin_freq_min.
  rewrite -> Nat.mod_small; omega.
  }
so (destruct_election _#6 Helection) as (Htf & Ht & Hf).
assert (f = 0).
  {
  destruct f as [| f]; auto.
  exfalso.
  so (weight_nonempty _#4 Hf (Nat.lt_0_succ _)) as (y & (Hwity & Hyw) & Hvote).
  exploit (Hunan y) as Hvote'.
    {
    split; auto.
    exists w.
    split; auto.
    reflexivity.
    }
  so (vote_fun _#5 Hvote Hvote') as H.
  discriminate H.
  }
subst f.
so (elector_minimum _#3 (Hwitw andel) Htf).
omega.
Qed.


Lemma famous_witness_exists_global :
  forall s i,
    exists x,
      fwitness s (global s) i x.
Proof.
intros s i.
so (all_rounds_inhabited _ (2 + first_regular + i) (global_nonterminating s)) as (y & Hwity & Wy).
so (famous_witness_exists s _ _ Hwity) as (x & Hwitx & Hdecision).
exists x.
split; auto.
exists y.
split; auto.
Qed.


Lemma no_true_votes_election :
  forall s i x y t f,
    (forall z, z @ y -> ~ vote s x z true)
    -> election (vote s x) i y t f
    -> t = 0.
Proof.
intros s i x y t f Hno Helection.
assert (weight creatorstake (fun z => elector i z y /\ vote s x z true) 0) as Hshutout.
  {
  apply weight_empty'.
  intros z (Helector & Hvote).
  apply (Hno z); auto.
  apply stsees_impl_strict_ancestor.
  exact (Helector ander).
  }
so (destruct_election _#6 Helection) as (_ & Ht & _).
eapply weight_fun; eauto.
Qed.


Lemma true_vote_impl_strict_ancestor :
  forall s x y,
    vote s x y true
    -> x @ y.
Proof.
intros s x y.
wfinduct y using strict_ancestor_well_founded.
intros y IH Hvote.
set (b := true) in Hvote.
assert (b = true) as Heq by auto.
clearbody b.
revert IH Heq.
cases Hvote.

(* first *)
{
intros s x y m n v _ _ _ _ H _ ->.
apply H.
cbn.
auto.
}

(* regular *)
{
intros s x y m n t f v _ Hwity Hn _ Helect H IH ->.
destruct H as [(Hle & _) | (_ & Heq)].
2:{
  discriminate Heq.
  }
so (destruct_election _#6 Helect) as (Hcard & Hcardt & _).
assert (t > 0) as Ht.
  {
  eassert _ as H; [refine (elector_minimum _#3 _ Hcard) |].
    {
    replace (S (pred n)) with n by omega.
    exact (Hwity andel).
    }
  omega.
  }
so (weight_nonempty _#4 Hcardt Ht) as (z & Hz & Hvote).
so (stsees_impl_strict_ancestor _ _ (Hz ander)) as Hzy.
eapply plus_trans; eauto.
apply IH; auto.
}

(* coin_super *)
{
intros s x y m n t f v _ Hwity Hn _ Helect Hgt IH ->.
so (destruct_election _#6 Helect) as (Hcard & Hcardt & _).
assert (t > 0) as Ht.
  {
  eassert _ as H; [refine (elector_minimum _#3 _ Hcard) |].
    {
    replace (S (pred n)) with n by omega.
    exact (Hwity andel).
    }
  omega.
  }
so (weight_nonempty _#4 Hcardt Ht) as (z & Hz & Hvote).
so (stsees_impl_strict_ancestor _ _ (Hz ander)) as Hzy.
eapply plus_trans; eauto.
apply IH; auto.
}

(* coin *)
{
intros s x y m n t f _ Hwity Hn _ Helect _ Hf IH _.
so (destruct_election _#6 Helect) as (Hcard & Hcardt & _).
assert (t > 0) as Ht.
  {
  eassert _ as H; [refine (elector_minimum _#3 _ Hcard) |].
    {
    replace (S (pred n)) with n by omega.
    exact (Hwity andel).
    }
  omega.
  }
so (weight_nonempty _#4 Hcardt Ht) as (z & Hz & Hvote).
so (stsees_impl_strict_ancestor _ _ (Hz ander)) as Hzy.
eapply plus_trans; eauto.
apply IH; auto.
}
Qed.


Lemma no_late_fame_strong :
  forall s i j x y,
    rwitness i x
    -> rwitness j y
    -> first_regular + i < j
    -> (j - i) mod coin_freq <> 0
    -> ~ x @= y
    -> decision s x y false.
Proof.
intros s i j x y Hwitx Hwity Hij Hnocoin Hnxy.
assert (forall z, elector (pred j) z y -> vote s x z false) as Hallno.
  {
  intros z Helector.
  so (elector_rwitness _#3 Helector) as Hwitz.
  eassert _ as H; [refine (vote_defined s _#4 Hwitx Hwitz _) |].
    {
    so first_regular_min.
    omega.
    }
  destruct H as (b & Hvote).
  destruct b; auto.
  exfalso.
  destruct Hnxy.
  so (true_vote_impl_strict_ancestor _#3 Hvote) as Hxz.
  apply plus_star.
  eapply plus_trans; eauto.
  eapply stsees_impl_strict_ancestor.
  exact (Helector ander).
  }
exploit (election_defined (vote s x) (pred j) y) as H.
  {
  intros w v v' Hvote Hvote'.
  eapply vote_fun; eauto.
  }

  {
  intros w Hwy Hwitw.
  refine (vote_defined s _#4 Hwitx Hwitw _).
  so first_regular_min.
  omega.
  }
destruct H as (t & f & Helection).
exists i, j, t, f.
do2 5 split; auto.
  {
  omega.
  }

  {
  so (destruct_election _#6 Helection) as (Hcard & Hcardt & _).
  cut (t = 0).
    {
    intro Heq.
    so (Hwity andel) as Hround.
    replace j with (S (pred j)) in Hround by omega.
    so (elector_minimum _#3 Hround Hcard).
    omega.
    }
  refine (weight_fun _#5 Hcardt _).
  apply weight_empty'.
  intros z (Helector & Hvote).
  so (Hallno _ Helector) as Hvote'.
  so (vote_fun _#5 Hvote Hvote') as H.
  discriminate H.
  }
Qed.


Lemma no_late_fame :
  forall s i x y w b,
    rwitness i x
    -> rwitness i y
    -> decision s x w b
    -> ~ y @= w
    -> decision s y w false.
Proof.
intros s i x y w b Hwitx Hwity Hdecision Hnyw.
destruct Hdecision as (i' & j & _ & _ & Hwitx' & Hwitw & Hlt & Hnocoin & _).
so (rwitness_fun _#3 Hwitx Hwitx'); subst i'; clear Hwitx'.
eapply no_late_fame_strong; eauto.
omega.
Qed.


(** Lemma 5.18 *)
Corollary no_late_fame_world :
  forall s W i x y w b,
    member W w
    -> ~ member W y
    -> rwitness i x
    -> rwitness i y
    -> decision s x w b
    -> decision s y w false.
Proof.
intros s W i x y w b Ww Wny Hwitx Hwity Hdecision.
eapply (no_late_fame _ _ x); eauto.
contradict Wny.
eapply world_closed; eauto.
Qed.


Lemma decision_ancestor :
  forall s x y,
    decision s x y true
    -> x @= y.
Proof.
intros s x y Hdecision.
so (ancestor_decide x y) as [| Hnxy]; auto.
exfalso.
so Hdecision as (i & _ & _ & _ & Hwitx & _).
so (no_late_fame _#6 Hwitx Hwitx Hdecision Hnxy) as Hdecision'.
so (decision_fun _#5 Hdecision Hdecision') as H.
discriminate H.
Qed.


Lemma famous_member :
  forall s W x,
    famous s W x
    -> member W x.
Proof.
intros s W x Hfamous.
destruct Hfamous as (y & Hsy & Hdecision).
eapply world_closed; eauto.
eapply decision_ancestor; eauto.
Qed.
